@@ -10,8 +10,11 @@ except ImportError:
     # Python < 3.7
     pkg_resources = None
 import random
+import sys
 from copy import copy
 from math import log2
+from time import sleep
+from typing import Tuple, Iterable
 
 from chessengine.moves import (
     get_white_pawn_moves,
@@ -34,7 +37,14 @@ from chessengine.lookup_tables import (
     pos_to_coords,
     san_piece_map,
 )
-from chessengine.utils import get_bit_positions, get_file, clear_lines
+from chessengine.utils import (
+    get_bit_positions,
+    get_file,
+    get_rank,
+    clear_lines,
+    get_input,
+    change_turn,
+)
 from chessengine.pgn.parser import PGNParser, SAN_MOVE_REGEX
 
 
@@ -42,12 +52,10 @@ class Board:
     """
     A class implementing a bitboard representation of a chess board.
     A particular bitboard can be accessed via the get_bitboard method or
-    as an attribute with the name <side>_<piece>s. For example -
+    as an attribute with the name <side>_<piece>s. For example, ``white_pawns``,
+    ``black_bishops``, ``white_queens``, ``black_kings``.
 
-        - white_pawns
-        - white_queens
-        - black_kings
-        - black_rooks
+    :param side: The side that the _board_ will play. Should be one of "white" or "black"
     """
 
     def __init__(self, side: str):
@@ -104,6 +112,11 @@ class Board:
             raise ValueError(f'side must be one of "black" or "white". Got {side}')
         self.side = side.lower().strip()
         self.opponent_side = "black" if self.side == "white" else "white"
+        self.en_passant_position = 0
+        self.white_king_side_castle = True
+        self.white_queen_side_castle = True
+        self.black_king_side_castle = True
+        self.black_queen_side_castle = True
 
         # A dictionary matching a side and piece to its corresponding bit board.
         # Useful when we want to iterate through all of the bitboards of the board.
@@ -126,7 +139,7 @@ class Board:
         self.moves = []
 
     def __repr__(self):
-        piece_list = ["\u2001" for _ in range(64)]
+        piece_list = [" " for _ in range(64)]
         unicode_piece = {
             ("white", "kings"): "\u2654",
             ("white", "queens"): "\u2655",
@@ -142,7 +155,7 @@ class Board:
             ("black", "pawns"): "\u265F",
         }
         ranks = ["8", "7", "6", "5", "4", "3", "2", "1"]
-        files = ["\u2001", "a", "b", "c", "d", "e", "f", "g", "h", "\u2001"]
+        files = [" ", "a", "b", "c", "d", "e", "f", "g", "h", " "]
 
         def add_bitboard_to_repr(board, s, p):
             board_string = bin(board)[2:]
@@ -154,18 +167,28 @@ class Board:
         for side, piece in self.board:
             add_bitboard_to_repr(self.board[(side, piece)], side, piece)
 
+        r = range(8)
+        if self.side == "white":
+            r = reversed(r)
+            files = list(reversed(files))
+
         board_repr = ""
-        board_repr += "\u2001" + "\u2001".join(files) + "\n"
-        for i in range(8):
-            board_repr += (
-                ranks[i]
-                + "\u2001\u2001"
-                + "\u2001".join(piece_list[8 * i : 8 * i + 8][::-1])
-                + "\u2001\u2001"
-                + ranks[i]
-            )
+        board_repr += " " + "   ".join(files) + "\n"
+
+        for i in r:
+            board_repr += "   +---+---+---+---+---+---+---+---+   \n"
+            board_repr += ranks[i] + "  | "
+
+            if self.side == "white":
+                board_repr += " | ".join(piece_list[8 * i : 8 * i + 8])
+            else:
+                board_repr += " | ".join(piece_list[8 * i : 8 * i + 8][::-1])
+
+            board_repr += " |  " + ranks[i]
             board_repr += "\n"
-        board_repr += "\u2001" + "\u2001".join(files) + "\n"
+
+        board_repr += "   +---+---+---+---+---+---+---+---+   \n"
+        board_repr += " " + "   ".join(files) + "\n"
         return board_repr
 
     def __str__(self):
@@ -177,8 +200,6 @@ class Board:
         return str(self) == str(other)
 
     def __hash__(self):
-        """Just hash the bitboards instead of the FEN
-        because it is faster and easier to maintain than the FEN state"""
         hash_str = f"{self.side[0]} "
         for side, piece in [
             ("white", "kings"),
@@ -199,6 +220,10 @@ class Board:
 
     @property
     def score(self):
+        """
+        The "score" of the board. A higher/more positive score favors white,
+        a lower/more negative score favors black.
+        """
         K = self.piece_count[("white", "kings")]
         Q = self.piece_count[("white", "queens")]
         R = self.piece_count[("white", "rooks")]
@@ -218,6 +243,9 @@ class Board:
 
     @property
     def board_pieces(self):
+        """
+        A list of all bitboards of the pieces that belong to the Board.
+        """
         if self.side == "white":
             return [
                 ("white", "kings"),
@@ -238,6 +266,9 @@ class Board:
 
     @property
     def opponent_pieces(self):
+        """
+        A list of all bitboards of the pieces that belong to the opponent.
+        """
         if self.side == "black":
             return [
                 ("white", "kings"),
@@ -258,13 +289,16 @@ class Board:
 
     def copy(self):
         """
-        Create and return a copy of self.
+        Create and return a copy of the board.
         """
         return copy(self)
 
     def get_side_bitboard(self, side: str) -> int:
         """
-        Returns the bitboard containing all pieces for the given side
+        Returns the bitboard containing all pieces for the given side.
+
+        :param side: "white" or "black"
+        :return: If ``side == "white"``, returns ``self.all_white``, else ``self.all_black``
         """
         if side == "white":
             return self.all_white
@@ -277,6 +311,10 @@ class Board:
 
         Raises AttributeError if a bitboard with an invalid name is requested. See above for the bitboard naming
         convention.
+
+        :param side: "white" or "black"
+        :param piece: Can be one of - "kings", "queens", "bishops", "knights", "rooks", "pawns"
+        :return: Bitboard
         """
         attrname = side + "_" + piece
         return getattr(self, attrname)
@@ -286,7 +324,8 @@ class Board:
         Returns the bitboard corresponding to the passed piece, considering the board's
         own side. i.e. - If the board is white, calling with piece="king" will return
         white king, etc.
-        piece can be one of - "kings", "queens", "bishops", "knights", "rooks", "pawns"
+        :param piece: Can be one of - "kings", "queens", "bishops", "knights", "rooks", "pawns"
+        :return: Bitboard
         """
         return self.get_bitboard(side=self.side, piece=piece)
 
@@ -294,7 +333,7 @@ class Board:
         """
         Updates all Board attributes when a move is made or the state of the Board
         changes in any way. You should call this if you manually make any changes to
-        Board attributes, otherwise it is called automatically.
+        bitboards attributes, otherwise it is called automatically.
         """
         self.all_white = (
             self.white_pawns
@@ -334,6 +373,10 @@ class Board:
     def set_bitboard(self, side: str, piece: str, board: int) -> None:
         """
         Sets the bitboard for the passed arguments to the passed bitboard
+
+        :param side: "white" or "black"
+        :param piece: Can be one of - "kings", "queens", "bishops", "knights", "rooks", "pawns"
+        :param board: The bitboard to be set
         """
         attrname = side + "_" + piece
         setattr(self, attrname, board)
@@ -341,13 +384,12 @@ class Board:
 
     def identify_piece_at(self, position: int) -> tuple:
         """
-        Identifies if there is any piece on the position passed. Position is a power of 2
+        Identifies if there is any piece on the position passed.
 
-        Returns a 3-tuple of the format (side, piece, bitboard) where side is the side of
-        the piece identified at position (e.g, "black"), piece is the type of piece identified
-        at position (e.g, "bishops"), and bitboard is the bitboard of the piece (e.g, Board.black_bishops).
-
-        If no piece is present at position, returns (None, None, None).
+        :param position: A power of 2 corresponding to a position on the board. See :ref:`position_representation`
+        :return: Returns a 3-tuple of the format (side, piece, bitboard) where side is the side of
+            the piece identified at position (e.g, "black"), piece is the type of piece identified
+            at position (e.g, "bishops"), and bitboard is the bitboard of the piece (e.g, Board.black_bishops).
         """
         for side, piece in self.board:
             board = self.board[(side, piece)]
@@ -358,10 +400,22 @@ class Board:
     def move(self, start: int, end: int, track: bool = True) -> None:
         """
         Moves the piece at start to end. Doesn't check anything, just makes
-        the move (unless the start or end positions are invalid).
+        the move (unless the start or end positions are invalid). Also checks if move
+        is a castle, moves the rook automatically on castle, and updates each
+        side's ability to castle on each move.
 
-        start and end are both powers of 2.
-        If track is True, stores the move in Board.moves so you can undo it later.
+        .. important::
+            This is the underlying function that is called by both ``Board.move_san`` and
+            ``Board.move_raw``. This function keeps track of castling status, but performs
+            no validation. ``Board.move_san`` and ``Board.move_raw`` perform validation,
+            but don't keep track of castling status. In general, you should use ``Board.move_san``
+            or ``Board.move_raw`` inside the game loop to ensure both move validation and
+            castling status are correctly performed/updated. Only use ``Board.move`` in special
+            cases when you want to make arbitrary moves outside the rules and/or the game loop.
+
+        :param start: The start position of the move. See :ref:`position_representation`
+        :param end: The end position of the move. See :ref:`position_representation`
+        :param track: If ``True``, the move made will be stored in self.moves
         """
         if not 1 <= start <= 2**63:
             raise ValueError(
@@ -382,11 +436,85 @@ class Board:
                 f"Can't move from {log2(start)} to {log2(end)}, both positions have {end_side} pieces."
             )
 
+        # Identify if the move is a castle and what type of castle it is
+        castle_type = None
+        # Don't set castling flags if we're not tracking this move
         if track:
-            # Keep track of the board state before the move was made so we can undo
-            start_state = (start, end, end_side, end_piece, end_board)
+            if start_piece == "kings":
+                if start_side == "white":
+                    self.white_king_side_castle = False
+                    self.white_queen_side_castle = False
+                    if start == 2**4:
+                        if end == 2**6:
+                            castle_type = "white_kingside"
+                        if end == 2**2:
+                            castle_type = "white_queenside"
+                if start_side == "black":
+                    self.black_king_side_castle = False
+                    self.black_queen_side_castle = False
+                    if start == 2**60:
+                        if end == 2**62:
+                            castle_type = "black_kingside"
+                        if end == 2**58:
+                            castle_type = "black_queenside"
+
+            # Set castling ability to false when rook is moved
+            if start_piece == "rooks":
+                if start_side == "white" and start == 1:
+                    self.white_queen_side_castle = False
+                elif start_side == "white" and start == 2**7:
+                    self.white_king_side_castle = False
+                elif start_side == "black" and start == 2**63:
+                    self.black_king_side_castle = False
+                elif start_side == "black" and start == 2**56:
+                    self.black_queen_side_castle = False
+
+        # Track moves made so we can undo
+        if track:
+            start_state = (start, end, end_side, end_piece, end_board, castle_type)
             self.moves.append(start_state)
 
+        # Check en passant moves
+        if start_piece == "pawns":
+            if start_side == "white":
+                # Check en passant status
+                if get_rank(end) - get_rank(start) == 2:
+                    self.en_passant_position = start << 8
+
+                # Check if a pawn captured by an en passant move
+                elif get_file(start) != get_file(end):
+                    # White pawn made an en passant move
+                    black_pawn_bb = self.get_bitboard("black", "pawns")
+                    black_pawn_bb &= clear_position[end >> 8]
+                    self.set_bitboard("black", "pawns", black_pawn_bb)
+                    self.piece_count[("black", "pawns")] -= 1
+                    self.en_passant_position = 0
+
+                # Clear self.en_passant_position
+                else:
+                    self.en_passant_position = 0
+
+            else:
+                # Check en passant status
+                if get_rank(start) - get_rank(end) == 2:
+                    self.en_passant_position = start >> 8
+
+                # Check if a pawn captured by an en passant move
+                elif get_file(start) != get_file(end):
+                    # Black pawn made an en passant move
+                    white_pawn_bb = self.get_bitboard("white", "pawns")
+                    white_pawn_bb &= clear_position[end << 8]
+                    self.set_bitboard("white", "pawns", white_pawn_bb)
+                    self.piece_count[("white", "pawns")] -= 1
+                    self.en_passant_position = 0
+
+                # Clear self.en_passant_position
+                else:
+                    self.en_passant_position = 0
+
+        else:
+            self.en_passant_position = 0
+
         if end_piece is not None:
             # Clear the captured piece's position (set "end" to 0)
             opp_side_board = self.get_bitboard(end_side, end_piece)
@@ -402,28 +530,93 @@ class Board:
         move_side_board |= mask_position[end]
         self.set_bitboard(start_side, start_piece, move_side_board)
 
+        # If the move was a castle, also move the rook into the correct position
+        # Validity of castling is not checked
+        if castle_type == "white_kingside":
+            self.move(2**7, 2**5, False)  # Don't track this move
+            self.white_king_side_castle = False
+            self.white_queen_side_castle = False
+        elif castle_type == "white_queenside":
+            self.move(2**0, 2**3, False)  # Don't track this move
+            self.white_king_side_castle = False
+            self.white_queen_side_castle = False
+        elif castle_type == "black_kingside":
+            self.move(2**63, 2**61, False)  # Don't track this move
+            self.black_king_side_castle = False
+            self.black_queen_side_castle = False
+        elif castle_type == "black_queenside":
+            self.move(2**56, 2**59, False)  # Don't track this move
+            self.black_king_side_castle = False
+            self.black_queen_side_castle = False
+
+    def move_raw(self, start: int, end: int, track: bool = True) -> None:
+        """
+        Moves the piece at start to end. Checks if the move is a valid move
+        to make given the current state of the board.
+
+        :param start: The start position of the move. See :ref:`position_representation`
+        :param end: The end position of the move. See :ref:`position_representation`
+        :param track: If ``True``, the move made will be stored in self.moves
+        """
+        side, piece, board = self.identify_piece_at(start)
+        if side is None:
+            raise ValueError(
+                f"There is no piece at {pos_to_coords[int(log2(start))]} to move."
+            )
+        moves = self.get_moves(side=side, piece=piece)
+        if (start, end) not in moves:
+            raise ValueError(
+                f"{pos_to_coords[int(log2(start))]} to {pos_to_coords[int(log2(end))]} is not a valid move for {side}"
+            )
+        self.move(start=start, end=end, track=track)
+
     def move_san(self, move: str, side: str) -> None:
         """
         Make a move given in standard algebraic notation.
-        move is the SAN move, side is the side to move.
+
+        :param move: A move given in SAN
+        :param side: "white" or "black"
         """
-        # TODO - Add support for undoing castling
         if "0-0-0" in move:
             # queen side castle
             if side == "white":
-                self.move(1, 2**3, track=False)
-                self.move(2**4, 2**2)
+                if self.white_queen_side_castle:
+                    self.move(2**4, 2**2)
+                else:
+                    raise ValueError(
+                        f'White cannot castle, it has already moved the {"rook" if self.white_king_side_castle else "king"}.'
+                    )
+                self.white_queen_side_castle = False
+                self.white_king_side_castle = False
             else:
-                self.move(2**56, 2**59, track=False)
-                self.move(2**60, 2**58)
+                if self.black_queen_side_castle:
+                    self.move(2**60, 2**58)
+                else:
+                    raise ValueError(
+                        f'Black cannot castle, it has already moved the {"rook" if self.black_king_side_castle else "king"}.'
+                    )
+                self.black_queen_side_castle = False
+                self.black_king_side_castle = False
         elif "0-0" in move:
             # king side castle
             if side == "white":
-                self.move(2**7, 2**5, track=False)
-                self.move(2**4, 2**6)
+                if self.white_king_side_castle:
+                    self.move(2**4, 2**6)
+                else:
+                    raise ValueError(
+                        f'White cannot castle, it has already moved the {"rook" if self.white_queen_side_castle else "king"}.'
+                    )
+                self.white_king_side_castle = False
+                self.white_queen_side_castle = False
             else:
-                self.move(2**63, 2**61, track=False)
-                self.move(2**60, 2**62)
+                if self.black_king_side_castle:
+                    self.move(2**60, 2**62)
+                else:
+                    raise ValueError(
+                        f'Black cannot castle, it has already moved the {"rook" if self.black_queen_side_castle else "king"}.'
+                    )
+                self.black_king_side_castle = False
+                self.black_queen_side_castle = False
         else:
             # regular move
             match = SAN_MOVE_REGEX.match(move)
@@ -440,25 +633,37 @@ class Board:
             moves = self.get_moves(side, piece_moved)
             if groups[1] is None:
                 # No file provided in the SAN
+                candidate_move = None
                 for m in moves:
                     if m[1] == end_pos:
-                        self.move(start=m[0], end=m[1])
-                        break
-                else:
+                        if candidate_move:
+                            raise ValueError(
+                                f"{move} is ambiguous for {side}. Specify a file to move from."
+                            )
+                        candidate_move = m
+                if candidate_move is None:
                     raise ValueError(f"{move} is not a valid move for {side}.")
+                else:
+                    self.move(start=candidate_move[0], end=candidate_move[1])
             else:
                 if groups[2] is None:
                     # No rank provided in the SAN
+                    candidate_move = None
                     for m in moves:
                         file = get_file(m[0])
                         if (
                             groups[1].upper() == "ABCDEFGH"[file - 1]
                             and m[1] == end_pos
                         ):
-                            self.move(start=m[0], end=m[1])
-                            break
-                    else:
+                            if candidate_move:
+                                raise ValueError(
+                                    f"{move} is ambiguous for {side}. Specify a file as well as rank to move from."
+                                )
+                            candidate_move = m
+                    if candidate_move is None:
                         raise ValueError(f"{move} is not a valid move for {side}.")
+                    else:
+                        self.move(start=candidate_move[0], end=candidate_move[1])
                 else:
                     # File and rank both present in the SAN
                     start_pos = 2 ** coords_to_pos[groups[1].upper() + groups[2]]
@@ -466,37 +671,65 @@ class Board:
                         raise ValueError(f"{move} is not a valid move for {side}.")
                     self.move(start=start_pos, end=end_pos)
 
-    def make_moves(self, *moves: tuple[int, int]) -> None:
+    def make_moves(self, *moves: Iterable[tuple[int, int]]) -> None:
         """
         Given a number of moves as tuples (start, end), call
-        Board.move on all
+        Board.move on all. Tracks all moves by default in ``self.moves``
         """
         for start, end in moves:
             self.move(start, end)
 
-    def undo_move(self):
+    def undo_move(self) -> None:
         """
-        Undo the last move tracked move.
+        Undo the last tracked move.
         """
         if not self.moves:
             raise RuntimeError("No moves have been made yet to undo.")
-        end, start, side, piece, board = self.moves.pop()
-        self.move(start=start, end=end, track=False)
-        if side is not None:
-            self.set_bitboard(side, piece, board)
-            self.piece_count[(side, piece)] += 1
+        end, start, side, piece, board, castle_type = self.moves.pop()
+
+        if castle_type is not None:
+            # TODO - if user castles when both self.white_kingside and self.white_queenside are
+            #       True, undoing this move only lets the user castle to the same side they castled
+            #       before undoing. Same for black.
+            if castle_type == "white_kingside":
+                self.move(start=start, end=end, track=False)  # Move king
+                self.move(2**5, 2**7)  # Move rook
+            elif castle_type == "white_queenside":
+                self.move(start=start, end=end, track=False)  # Move king
+                self.move(2**3, 2**0)  # Move rook
+            elif castle_type == "black_kingside":
+                self.move(start=start, end=end, track=False)  # Move king
+                self.move(2**61, 2**63)  # Move rook
+            elif castle_type == "black_queenside":
+                self.move(start=start, end=end, track=False)  # Move king
+                self.move(2**59, 2**56)  # Move rook
+        else:
+            self.move(start=start, end=end, track=False)
+            if side is not None:
+                self.set_bitboard(side, piece, board)
+                self.piece_count[(side, piece)] += 1
 
     def get_moves(
         self, side: str, piece: str = None, position: int = None
     ) -> list[tuple[int, int]]:
         """
         Get all end positions a piece of side can reach starting from position.
-        side is always required, piece and position are optional.
-        If piece is not specified, get all moves for all pieces of the passed side, i.e. get all valid moves for white or black.
-        If side and piece are specified and position is not, get all valid moves for the specified side and piece on
-        the board, i.e. if side is "white" and piece is "rooks", get all valid moves for all white rooks on the board.
+        ``side`` is always required, piece and position are optional.
 
-        Returns a list of moves as a list of tuples (start, end), where start and end are positions on the board.
+        If piece is not specified, gets all moves for all pieces of the passed side, i.e. get
+        all valid moves for white or black.
+
+        If side and piece are specified and position is not, gets all valid moves for the
+        specified side and piece on the board, i.e. if side is "white" and piece is "rooks",
+        gets all valid moves for all white rooks on the board.
+
+        If side, piece, and position are specified, gets all moves for the piece present on position.
+
+        :param side: "white" or "black"
+        :param piece: Can be one of - "kings", "queens", "bishops", "knights", "rooks", "pawns"
+        :param position: A power of 2 corresponding to a position on the board. See :ref:`position_representation`
+        :return: A list of moves as tuples ``(start, end)``, where ``start`` and ``end`` are positions on the board.
+            See :ref:`position_representation`
         """
         if piece is not None:
             if position is None:
@@ -534,13 +767,14 @@ class Board:
                     moves.extend(self.get_moves(side, piece, position))
             return moves
 
-    def search_forward(self, depth: int = 4) -> tuple[int, tuple]:
+    def search_forward(self, depth: int = 4) -> tuple[int, tuple[int, int]]:
         """
         Execute an alpha-beta pruned depth-first search to find the optimal move from
         the current board state.
 
-        Arguments -
-        depth: int - The number of plies to search (1 move is 2 plies). Default = 4 plies.
+        :param depth: int - The number of plies to search (1 move is 2 plies). Default = 4 plies.
+        :return: A 2-tuple where the first element is the best board score found, and the second
+            element is the best found move as a 2 tuple containing the start position and the end position
         """
         maximize = self.side == "white"
         best_score = -1000 if maximize else 1000
@@ -579,7 +813,7 @@ class Board:
         :param beta: The maximum score that the minimizing player is guaranteed (default=1000). You probably won't need to specify this argument.
         :param maximizing_player: True if white is searching for a move, False if black is searching for a move.
 
-        Returns the value of the best board position found.
+        :return: The score of the best board position found.
         """
         if depth == 0:
             return self.score
@@ -609,9 +843,68 @@ class Board:
                 beta = min(beta, value)
             return value
 
+    def handle_player_move(
+        self, side_to_move: str, last_move: str
+    ) -> Tuple[str, int, bool]:
+        """
+        Ask for user input until accepted. When a valid input is received, execute the move.
+
+        :param side_to_move: "white" or "black"
+        :param last_move: The last move made by the board to print out to the user
+        :return: A 3-tuple of the form - the valid (move, lines_printed, move_undone)
+            where move is the valid move received as input, lines_printed is the number
+            of lines printed, and move_undone is a boolean that is ``True`` when the user
+            chose to undo their previous move
+        """
+        move = ""
+        lines_added = 0
+        input_accepted = False
+
+        while not input_accepted:
+            print(last_move)
+            move = get_input(
+                f"Enter your move in standard algebraic notation ({side_to_move.capitalize()}'s turn) - "
+            )
+            lines_added += 2
+            if move.lower() == "q":
+                print("Thanks for playing!")
+                sys.exit(0)
+            if move.lower() == "u":
+                try:
+                    self.undo_move()
+                    self.undo_move()  # Undo both player's moves
+                except RuntimeError:
+                    print("No moves have been made yet to undo!")
+                    lines_added += 1
+                return move, lines_added, True
+            # Input was normal move
+            try:
+                if " to " in move:
+                    # Input was not in SAN
+                    _ = move.upper().strip().split()
+                    start = 2 ** coords_to_pos[_[0]]
+                    end = 2 ** coords_to_pos[_[-1]]
+                    self.move_raw(start=start, end=end)
+                else:
+                    # Input was in SAN
+                    self.move_san(move=move, side=side_to_move)
+                input_accepted = True
+                last_move = f"{side_to_move.capitalize()} moved {move}"
+            except ValueError as e:
+                print(e)
+                lines_added += 1
+            except KeyError as e:
+                print(f"{e} is not a valid square. Please try again.")
+                lines_added += 1
+        return move, lines_added, False
+
     def play(self, search_depth: int = 4) -> None:
         """
-        The game loop.
+        Play a game of chess against the computer.
+
+        :param search_depth: The number of plies the computer should search forward. Be careful
+            passing values about 4 as the search depth. It increases the running time of the
+            move search exponentially.
         """
         parser = None
         if pkg_resources is not None:
@@ -619,7 +912,7 @@ class Board:
                 "Searching for opening moves.",
                 "Reading an opening book.",
                 "Waking up.",
-                "Reading thousands of past games.",
+                "Building move tree.",
             ]
 
             parser = PGNParser()
@@ -629,21 +922,23 @@ class Board:
                 file_path = pkg_resources.path("chessengine.openings", child)
                 with file_path as f:
                     if f.suffix == ".pgn":
-                        print(".")
+                        print(".", end="", flush=True)
                         parser.parse(f)
             print(f"\nRead through {len(parser.games)} games.")
+        sleep(1)
         print(f"Set search depth to {search_depth}")
+        sleep(0.7)
 
         print(self)
         side_to_move = "white"
         in_game_tree = parser is not None
         current_node = parser.root_node if in_game_tree else None
-        lines_printed = 11
+        lines_printed = 20
         last_move = ""
         while True:
             clear_lines(lines_printed)
             print(self)
-            lines_printed = 11
+            lines_printed = 20
             if side_to_move == self.side:
                 if in_game_tree:
                     move, node = random.choice(list(current_node.children.items()))
@@ -655,40 +950,15 @@ class Board:
                     self.move(best_move[0], best_move[1])
                     last_move = f"Board moves from {pos_to_coords[log2(best_move[0])]} to {pos_to_coords[log2(best_move[1])]}"
             else:
-                # ask for user input until accepted:
-                input_accepted = False
-                move_undone = False
-                move = ""
-                while not input_accepted:
-                    print(last_move)
-                    move = input(
-                        "Enter the move you want to make in standard algebraic notation - "
-                    ).strip()
-                    lines_printed += 2
-                    if move.lower() == "q":
-                        print("Thanks for playing!")
-                        return
-                    if move.lower() == "u":
-                        move_undone = True  # this helps us break out of the input loop
-                        try:
-                            self.undo_move()
-                            self.undo_move()
-                        except RuntimeError:
-                            print("No moves have been made yet to undo!")
-                            lines_printed += 1
-                        break
+                move, lines_added, move_undone = self.handle_player_move(
+                    side_to_move, last_move
+                )
+                lines_printed += lines_added
+                last_move = f"{side_to_move.capitalize()} moved {move}"
 
-                    # input was normal move
-                    try:
-                        self.move_san(move=move, side=side_to_move)
-                        input_accepted = True
-                    except ValueError as e:
-                        print(e)
-                        lines_printed += 1
                 if move_undone:
                     # return to outer loop, so both sides need to make a new move
                     continue
-
                 if in_game_tree:
                     try:
                         current_node = current_node.get_child(move)
@@ -697,53 +967,31 @@ class Board:
                 print(f"You moved {move}")
                 lines_printed += 1
 
-            if side_to_move == "white":
-                side_to_move = "black"
-            else:
-                side_to_move = "white"
+            side_to_move = change_turn(side_to_move)
 
     def play_pvp(self) -> None:
         """
         Play a game of chess on the terminal with another player
         """
         print(self)
-        lines_printed = 11
+        lines_printed = 20
 
         side_to_move = "white"
         last_move = ""
         while True:
             clear_lines(lines_printed)
             print(self)
-            lines_printed = 11
+            lines_printed = 20
 
-            input_accepted = False
-            while not input_accepted:
-                print(last_move)
-                move = input(
-                    f"Enter the move you want to make ({side_to_move.capitalize()}'s turn) - "
-                ).strip()
-                lines_printed += 2
+            move, lines_added, move_undone = self.handle_player_move(
+                side_to_move, last_move
+            )
+            lines_printed += lines_added
+            last_move = f"{side_to_move.capitalize()} moved {move}"
 
-                if move.lower() == "q":
-                    print("Thanks for playing!")
-                    return
-                if move.lower() == "u":
-                    try:
-                        self.undo_move()
-                    except RuntimeError:
-                        print("No moves have been made yet to undo!")
-                        lines_printed += 1
-                    break
+            if (
+                move_undone
+            ):  # return to outer loop, so both sides need to make a new move
+                continue
 
-                try:
-                    self.move_san(move=move, side=side_to_move)
-                    input_accepted = True
-                    last_move = f"{side_to_move.capitalize()} moved {move}"
-                except ValueError as e:
-                    print(e)
-                    lines_printed += 1
-
-            if side_to_move == "white":
-                side_to_move = "black"
-            else:
-                side_to_move = "white"
+            side_to_move = change_turn(side_to_move)
