@@ -17,7 +17,15 @@ class Game:
         self.result = ""
         self.move_text = ""
 
-    def add_header(self, key, value):
+    def add_header(self, key: str, value: str) -> None:
+        """
+        Add a header to Game.headers
+
+        :param key: The header to be added (usually sourced from the PGN file)
+        :param value: The value of the header
+
+        :raises ValueError: If the key has already been added to the Game
+        """
         if key in self.headers:
             raise ValueError(
                 f"{key} header has already been set on this game - {self.headers[key]}."
@@ -30,6 +38,11 @@ class GameNode:
     A class representing a node in a game. Useful for
     parsing a PGN game and building a tree of moves
     for the opening book.
+
+    Provides a ``children`` dictionary that maps a move (str) made from
+    this GameNode to the GameNode representing the new Board state.
+
+    :param turn: The side to move on this GameNode.
     """
 
     def __init__(self, turn: str) -> None:
@@ -46,6 +59,13 @@ class GameNode:
         return move in self.children
 
     def add_child(self, move: str):
+        """
+        Create a new GameNode and add it as a child to the current node in
+        the ``children`` dictionary.
+
+        :param move: The move made to reach the new game node
+        :return GameNode: Return the newly created GameNode.
+        """
         if move not in self.children:
             turn = "white" if self.turn == "black" else "black"
             new_node = GameNode(turn)
@@ -54,6 +74,14 @@ class GameNode:
         return self.children[move]
 
     def get_child(self, move: str):
+        """
+        Check if ``move`` is a child of this GameNode. If it is,
+        return the corresponding GameNode, else raise ValueError.
+
+        :param move: The move to check for
+        :return GameNode: Return the GameNode if found
+        :raises ValueError: If the move passed is not a child of the GameNode
+        """
         if move not in self.children:
             raise ValueError(f"{move} is not a child of the current Game Node.")
         return self.children[move]
